package com.cringe_studios.cringe_authenticator_library.impl;

import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

import com.cringe_studios.cringe_authenticator_library.OTP;
import com.cringe_studios.cringe_authenticator_library.OTPAlgorithm;
import com.cringe_studios.cringe_authenticator_library.OTPException;
import com.cringe_studios.cringe_authenticator_library.OTPType;

//https://datatracker.ietf.org/doc/html/rfc4226
public class HOTP extends OTP {

	//protected static final int[] DIGITS_POWER
	// 0 1 2 3 4 5 6 7 8
	//		= { 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000 };

	public HOTP(String secret, OTPAlgorithm algorithm, int digits, long counter, long periodInSeconds, boolean checksum) {
		super(OTPType.HOTP, secret, algorithm, digits, counter, periodInSeconds, checksum);
	}

	@Override
	public long getNextDueTime() {
		return 0;
	}

	@Override
	public String getPin() throws OTPException{
		try {
			long movingFactor = this.getCounter();
			
			byte[] text = new byte[Long.BYTES];
			for (int i = text.length - 1; i >= 0; i--) {
				text[i] = (byte) (movingFactor & 0xff);
				movingFactor >>= 8;
			}
			
			return generateOTP(-1, text);
			
		} catch (InvalidKeyException e) {
			throw new OTPException("Your secret is invalid! Please rescan the code!", e);
		} catch (NoSuchAlgorithmException e) {
			throw new OTPException("Your platform does not support the " + this.getAlgorithm().name() + "algorithm!", e);
		}
	}

	public String generateOTP(int truncationOffset, byte[] message) throws NoSuchAlgorithmException, InvalidKeyException {
		boolean addChecksum = this.hasChecksum();
		byte[] secret = this.secret;
		int codeDigits = this.getDigits();

		// put movingFactor value into text byte array
		
		int digits = addChecksum ? (codeDigits + 1) : codeDigits;
		byte[] text = message;
		

		// compute hmac hash
		byte[] hash = this.getAlgorithm().hash(secret, text);

		// put selected bytes into result int
		int offset = hash[hash.length - 1] & 0xf;
		if ((0 <= truncationOffset) && (truncationOffset < (hash.length - 4))) {
			offset = truncationOffset;
		}
		int binary = ((hash[offset] & 0x7f) << 24) | ((hash[offset + 1] & 0xff) << 16)
				| ((hash[offset + 2] & 0xff) << 8) | (hash[offset + 3] & 0xff);

		//int otp = binary % DIGITS_POWER[codeDigits];
		
		//byte[] truncatedHash = Arrays.copyOfRange(hash, offset, offset + 3);
		//truncatedHash[0] &= 0x7F;
		//BigInteger bigBinary = new BigInteger(1, truncatedHash);
		BigInteger bigBinary = BigInteger.valueOf(binary);
		BigInteger bigOtp = bigBinary.mod(BigInteger.TEN.pow(codeDigits));
		
		//System.out.println(binary + " vs. " + bigBinary.toString());
		//System.out.println(otp + " vs. " + bigOtp.toString());
		
		if (addChecksum) {
			//otp = (otp * 10) + HOTPChecksumProvider.calcChecksum(otp, codeDigits);
			bigOtp = bigOtp.multiply(BigInteger.TEN).add(BigInteger.valueOf(HOTPChecksumProvider.calcChecksum(bigOtp.longValue(), codeDigits)));
		}
		//String result = Integer.toString(otp);
		String result = bigOtp.toString();
		while (result.length() < digits) {
			result = "0" + result;
		}
		return result;
	}

	@Override
	public void incrementCounter() {
		this.counter++;
	}

}
