package com.cringe_studios.cringe_authenticator_library.impl;

import java.nio.ByteBuffer;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import com.cringe_studios.cringe_authenticator_library.OTPAlgorithm;
import com.cringe_studios.cringe_authenticator_library.OTPException;
import com.cringe_studios.cringe_authenticator_library.OTPType;

//https://datatracker.ietf.org/doc/html/rfc6238
public class TOTP extends HOTP {
	
	public TOTP(String secret, OTPAlgorithm algorithm, int digits, long counter, long periodInSeconds, boolean checksum) {
		super(secret, algorithm, digits, counter, periodInSeconds, checksum);
		this.type = OTPType.TOTP;
	}
	
	@Override
	public long getNextDueTime() {
		// e.g. current time: 20 seconds & current period: 30 seconds -> next due time:
		// 30 seconds
		return (this.getCounter() + 1) * this.getPeriod();
	}

	@Override
	public String getPin() throws OTPException{
		try {
			return getPinAt(System.currentTimeMillis() / 1000);
		} catch (InvalidKeyException e) {
			throw new OTPException("Your secret is invalid! Please rescan the code!", e);
		} catch (NoSuchAlgorithmException e) {
			throw new OTPException("Your platform does not support the " + this.getAlgorithm().name() + "algorithm!", e);
		}
	}

	@Override
	public long getCounter() {
		return getCounterAt(System.currentTimeMillis() / 1000 - super.getTimeCorrection());
	}
	
	private long getCounterAt(long unixSecond) {
		return unixSecond / this.getPeriod();
	}

	public String getPinAt(long time) throws InvalidKeyException, NoSuchAlgorithmException {
		//int codeDigits = this.getDigits();

		// Get the HEX in a Byte[]
		byte[] msg = ByteBuffer.allocate(Long.BYTES).putLong(this.getCounterAt(time)).array();
		//byte[] k = this.secret;
		//byte[] hash = this.getAlgorithm().hash(k, msg);

		// put selected bytes into result int
		//int offset = hash[hash.length - 1] & 0xf;

		//int binary = ((hash[offset] & 0x7f) << 24) | ((hash[offset + 1] & 0xff) << 16)
		//		| ((hash[offset + 2] & 0xff) << 8) | (hash[offset + 3] & 0xff);

		//int otp = binary % DIGITS_POWER[codeDigits];

		//String result = Integer.toString(otp);
		//while (result.length() < codeDigits) {
		//	result = "0" + result;
		//}
		//return result;
		
		return this.generateOTP(-1, msg);
	}
	
	@Override
	public void incrementCounter() {}
}